<copyright> Random classes.
    Written by <a href="mailto:tiggr@ics.ele.tue.nl">Pieter J. Schoenmakers</a>

    Copyright &copy; 1996, 1997 Pieter J. Schoenmakers.

    This file is part of TOM.  TOM is distributed under the terms of the
    TOM License, a copy of which can be found in the TOM distribution; see
    the file LICENSE.

    Some of the algorithms used in these classes are described in
    ``Numerical Recipies'' by Press et al.

    <id>$Id: Random.t,v 1.4 2000/08/16 21:11:03 tiggr Exp $</id>
    </copyright>

implementation class
Random: instance (All)

end;

implementation instance
Random: instance (All)

<doc> Return a double random number.    </doc>
deferred double
  next;

<doc> Return an integer random number.  </doc>
deferred int
  next;

// This postcondition is a clear example of being elegant because of its
// documentational clarity, but which also never needs to be checked since
// it it so obviously always true.
// Wed Apr  9 01:35:52 1997, tiggr@tricky.es.ele.tue.nl
<doc> Return a number in the range [0, limit).  </doc>
int (result)
  next int limit
pre
  limit > 1
post
  result >= 0 && result < limit
{
  = [self next] % limit;
}

end;

/******************** IntegerRandom ********************/

<doc> Superclass for use by random number generators which customarily
    return ints.  </doc>
implementation class
IntegerRandom: Random, Limits

end;

implementation instance
IntegerRandom

<doc> Return the next double value, in the range [0, 1.0).  </doc>
double
  next
{
  int n = [self next];

  = double (n) / double (INT_MAX);
}

end;

/******************** DoubleRandom ********************/

<doc> Superclass for use by random number generators which customarily
    return ints.  </doc>
implementation class
DoubleRandom: Random, Limits

end;

implementation instance
DoubleRandom

<doc> Return the next integer value, in the range [0, INT_MAX].  </doc>
int (result)
  next
{
  double n = [self next];

  result = int (n * double (INT_MAX));

  /* Protect against a double-number generator which generates values
     outside [0, 1].  */
  if (result < 0)
    result = -result;
}

end;

/******************** MinimalRandom ********************/

<doc> MinimalRandom implements {IntegerRandom}.  It uses seeds 
    to generate a repeatable sequence of pseudo-random integers.  </doc>
implementation class
MinimalRandom: State, IntegerRandom
{
  <doc> Constants needed by the algorithm.  </doc>
  const ia = 16807;
  const im = 2147483647;
  const iq = 12773;
  const ir = 2836;
}

end;

implementation instance
MinimalRandom
{
  // Does a 32-bit seed suffice in any way?
  // Wed Apr  9 01:12:20 1997, tiggr@tricky.es.ele.tue.nl
  <doc> The seed from which we feed.  </doc>
  int seed;
}

<doc> Designated initializer.  </doc>
id (self)
  init int initial_seed
pre
  initial_seed > 0
{
  seed = initial_seed;
}

<doc> Initialize this instance with a seed derived from the current moment
    in time and the {hashq} value of {self}.  </doc>
id
  init
{
  int seed;

  do
    {
      // CCC `2e8' is lexed as an int...
      // Wed Apr  9 02:00:00 1997, tiggr@tricky.es.ele.tue.nl
      seed = int (2.0e8 * [Date relativeTimeIntervalSinceNow]) ^ [self hashq];

      if (seed < 0)
	seed = -seed - 1;
    } while (!seed);

  = [self init seed];
}

<doc> Return the next random positive {int} value.  </doc>
int
  next
{
  int i = seed / iq;

  seed = ia * (seed - i * iq) - i * ir;
  if (seed < 0)
    seed += im;

  = seed;
}

end;
