/* TOM/TRT operating system dependencies.
   Written by Pieter J. Schoenmakers <tiggr@gerbil.org>

   Copyright (C) 1996-2001 Pieter J. Schoenmakers.

   This file is part of TOM.  TOM is distributed under the terms of the
   TOM License, a copy of which can be found in the TOM distribution; see
   the file LICENSE.

   $Id: os.c,v 1.18 2001/04/21 08:01:23 bruce Exp $  */

#include "trt.h"
#include <tom/tom-r.h>

#if HAVE_OS_OS_C
#include <os/os.c>
#endif

#if HAVE_DLFCN_H
#define _GNU_SOURCE
#include <dlfcn.h>

void
dload_initialize ()
{
}

char *
dload_error (void)
{
  return (void *) dlerror ();
}

void *
dload_load (char *name)
{
  return dlopen (name, RTLD_NOW | RTLD_GLOBAL);
}

void *
dload_lookup (void *handle, char *name)
{
  return dlsym (handle, name);
}

#ifdef HAVE_DLADDR
void
dload_symbol_info (void *address,
		   tom_object *file_name,
		   void **base_address,
		   tom_object *symbol_name,
		   void **symbol_address,
		   void **offset)
{
  Dl_info sym_info;
  int ret;

  ret = dladdr(address, &sym_info);
  if (ret)
    {
      /* The casts to char* are to elminate const-ness.  I hope that
       * byte_string_with_c_string() doesn't modify that pointer!
       *  -- Bruce <bruce@cubik.org>, May 5, 2000.
       */
      *file_name = byte_string_with_c_string((char*)sym_info.dli_fname);
      *base_address = sym_info.dli_fbase;
      *symbol_name = byte_string_with_c_string((char*)sym_info.dli_sname);
      *symbol_address = sym_info.dli_saddr;
      *offset = (void*)((long)address - (long)*symbol_address);
    }
  else
    {
      *file_name = byte_string_with_c_string("????");
      *base_address = NULL;
      *symbol_name = byte_string_with_c_string("????");
      *symbol_address = NULL;
      *offset = NULL;
    }
}
#else /* HAVE_DLADDR */
void
dload_symbol_info (void *address,
                   tom_object *file_name,
                   void **base_address,
                   tom_object *symbol_name,
                   void **symbol_address,
                   void **offset)
{
    *file_name = byte_string_with_c_string("????");
    *base_address = NULL;
    *symbol_name = byte_string_with_c_string("????");
    *symbol_address = NULL;
    *offset = NULL;
}
#endif /* HAVE_DLADDR */
#endif /* HAVE_DLFCN_H */

#if HAVE_BACKTRACE

/* This is specific to glibc 2.1 and later.  */
#include <execinfo.h>

/* Fill at most the first LENGTH slots in PTRARRAY with outstanding return
   addresses.  Return the number of slots in actually filled.  */
int
trt_crawl_stack (void ***ptrarray, int length)
{
  return backtrace (*ptrarray, length);
}

#else

int
trt_crawl_stack (void ***ptrarray, int length)
{
  return 0;
}

#endif
